import json
from datetime import date, datetime


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat().split(".")[0]
        try:
            return super(JSONEncoder, self).default(obj)
        except TypeError:
            return str(obj)  # fallback to unicode


def json_dumps(body):
    return json.dumps(body, cls=JSONEncoder)



import pytest

class TestUtils(object):

    def test_maybe_dotted(self):
        from slovar.utils import maybe_dotted

        with pytest.raises(ImportError):
            maybe_dotted('aa')

        with pytest.raises(ValueError):
            maybe_dotted('')

        with pytest.raises(ValueError):
            maybe_dotted('.view')

        import slovar
        assert maybe_dotted('slovar.utils') == slovar.utils
        assert maybe_dotted('slovar.utils:maybe_dotted') == slovar.utils.maybe_dotted

        maybe_dotted('XYZ', throw=False)



import six
from urllib.parse import parse_qsl

from slovar.strings import split_strip, str2dt, str2rdt


def parametrize(func):

    def wrapper(dset, name, default=None, raise_on_values=None, pop=False,
                            allow_missing=False, set_as=None, pop_empty=False,
                            _raise=True, mod=None, **kw):

        if pop_empty:
            allow_missing = True

        if default is None:
            try:
                value = dset[name]
            except KeyError:
                if not allow_missing:
                    raise KeyError("Missing '%s'" % name)
                else:
                    return
        else:
            value = dset.get(name, default)

        if pop_empty and not value:
            dset.pop(name, None)
            return

        if raise_on_values and value in raise_on_values:
            raise ValueError("'%s' can not be any of %s" % (name, raise_on_values))

        try:
            result = func(dset, value, **kw)
        except Exception:
            if _raise:
                import sys
                raise ValueError(sys.exc_info()[1])
            else:
                return

        if mod:
            result = mod(result)

        if pop:
            dset.pop(name, None)
            if set_as:
                dset[set_as] = result
        else:
            dset[set_as or name] = result

        return result

    return wrapper


@parametrize
def asbool(dset, value):
    truthy = frozenset(('t', 'true', 'y', 'yes', 'on', '1'))
    falsey = frozenset(('f', 'false', 'n', 'no', 'off', '0'))

    if value is None:
        return False

    if isinstance(value, bool):
        return value

    lvalue = str(value).strip().lower()
    if lvalue in truthy:
        return True
    elif lvalue in falsey:
        return False
    else:
        raise ValueError(
                'Dont know how to convert `%s` to bool' % value)


@parametrize
def aslist(dset, value, sep=',', remove_empty=True, unique=False, itype=None):
    if isinstance(value, list):
        _lst = value
    elif isinstance(value, str):
        _lst = split_strip(value, sep, remove_empty)
    else:
        _lst = [value]

    if remove_empty:
        _lst = [it for it in _lst if it is not None]

    if unique:
        _lst = list(set(_lst))

    if itype:
        _lst = [itype(it) for it in _lst]

    return _lst


def asset(*args, **kw):
    return set(aslist(*args, **kw))


@parametrize
def asint(dset, value):
    return int(value)


@parametrize
def asfloat(dset, value):
    return float(value)


@parametrize
def asstr(dset, value):
    return str(value)


@parametrize
def asunicode(dset, value):
    return str(value)


@parametrize
def asrange(dset, value, typecast=str, sep='-'):
    if isinstance(value, list):
        list_ = value
    elif isinstance(value, str):
        rng = split_strip(value, sep)
        if len(rng) != 2:
            raise ValueError('bad range')
        list_ = list(range(int(rng[0]), int(rng[1])+1))

    else:
        list_ = [value]

    return [typecast(e) for e in list_]


def asdict(dset, name, _type=None, _set=False, pop=False):
    """
    Turn this 'a:2,b:blabla,c:True,a:'d' to {a:[2, 'd'], b:'blabla', c:True}

    """

    try:
        value = dset[name]
    except KeyError:
        raise KeyError("Missing '%s'" % name)

    if _type is None:
        _type = lambda t: t

    dict_str = dset.pop(name, None)
    if not dict_str:
        return {}

    _dict = {}
    for item in split_strip(dict_str):
        key, _, val = item.partition(':')
        if key in _dict:
            if type(_dict[key]) is list:
                _dict[key].append(val)
            else:
                _dict[key] = [_dict[key], val]
        else:
            _dict[key] = _type(val)

    if _set:
        dset[name] = _dict
    elif pop:
        dset.pop(name, None)

    return _dict


@parametrize
def asdt(dset, value):
    return str2dt(value)


def qs2dict(qs):
    from slovar import slovar
    return slovar(parse_qsl(qs, keep_blank_values=True))


@parametrize
def asqs(dset, value):
    return qs2dict(value)

from bson import ObjectId
@parametrize
def asdtob(dset, value):
    return str(ObjectId.from_datetime(str2dt(value)))



from slovar.strings import split_strip


def expand_list(param):
    _new = []
    if isinstance(param, (list, set)):
        for each in param:
            if isinstance(each, str) and each.find(',') != -1:
                _new.extend(split_strip(each))
            elif isinstance(each, (list, set)):
                _new.extend(each)
            else:
                _new.append(each)
    elif isinstance(param, str) and param.find(',') != -1:

        _new = split_strip(param)

    return _new


def process_fields(fields):
    # Avoid circular dependencies
    from slovar import slovar

    fields_only = []
    fields_exclude = []
    show_as = {}
    show_as_r = {}
    transforms = {}
    assignments = {}
    flats = {}
    unflats = []
    envelope = None
    star = False
    exclude_field = False
    exp_only = []

    if isinstance(fields, str):
        fields = split_strip(fields)

    for field in expand_list(fields):
        field = field.strip()

        if not field:
            continue

        if '*' == field:
            star = True
            continue

        if ':=' in field:
            kk, _, val = field.partition(':=')
            assignments[kk] = val
            # continue

        field,_,trans = field.partition(':')
        trans = trans.split('|') if trans else []

        if field[0] == '-':
            field = field[1:]
            exclude_field = True

        if '__as__' in field:
            root,_,val = field.partition('__as__')
            if not root and val:
                envelope = val
                continue
            else:
                if root in show_as:
                    show_as[root].append(val or root.split('.')[-1])
                else:
                    show_as[root] = [val or root.split('.')[-1]]

                show_as_r[val or root.split('.')[-1]]=root
                field = root

        if trans:
            if field in show_as:
                tr_field = show_as[field][0]
            else:
                tr_field = field

            if 'flat' in trans:
                flats[tr_field] = 1
                trans.remove('flat')

            elif 'flatall' in trans:
                flats[tr_field] = 0
                trans.remove('flatall')

            if 'unflat' in trans:
                trans.remove('unflat')
                unflats.append(tr_field)

            transforms[tr_field] = trans

        if field in unflats:
            continue

        if '.' in field:
            root = field.split('.')[0]
        else:
            root = field

        if exclude_field:
            fields_exclude.append(root)
            exclude_field=False
        else:
            fields_only.append(root)
            exp_only.append(field)


    return slovar({
             'fields': fields,
             'only': fields_only,
             'exclude':fields_exclude,
             'show_as': show_as,
             'show_as_r': show_as_r,
             'transforms': transforms,
             'assignments': assignments,
             'star': star,
             'flats': flats,
             'unflats': unflats,
             'envelope': envelope,
             'exp_only': exp_only,
             })


def union_fields(f1, f2):
    f1 = process_fields(f1)
    f2 = process_fields(f2)

    'a,b__as__bb,c.d,e.f__as__g,*'
    'a,x,c'


def sort_list(items, by='', reverse=False):
    'sort generic list of basic type or nested dicts'

    _items = []
    none_items = []

    if not by:
        return sorted(items, reverse=reverse)

    for each in items:
        if isinstance(each, dict):
            val = each.get(by)

            if val is None:
                none_items.append(each)
                continue

            _items.append(each)

    sorted_list = sorted(_items, key=lambda x: x.get(by), reverse=reverse)

    if reverse:
        return sorted_list + none_items
    else:
        return none_items + sorted_list









import re
import logging
import collections
import logging
import builtins
import copy
from bson import ObjectId
from datetime import datetime

from slovar import convert
from slovar.dictionaries import *
from slovar.json import json_dumps
from slovar.lists import *
from slovar.strings import *


def ld2l(ld, key):
    return [it[key] for it in ld]


def ld2dl(ld, key=None):
    dl = {}
    for _d in ld:
        for kk,vv in _d.items():
            if kk in dl:
                dl[kk].append(vv)
            else:
                dl[kk] = [vv]
    return dl


def parse_func_params(s):
    pattern = r'(\w[\w\d_]*)\((.*)\)$'
    match = re.match(pattern, s)
    if match:
        return list(match.groups())
    else:
        return []

TCAST_NONE = True
TCAST_FUNCS = ['sort', 'index', 'concat', 'slice', 'ld2l', 'split']

log = logging.getLogger(__name__)


class slovar(dict):
    """Named dict, with some set functionalities
    """

    @classmethod
    def to_dicts(cls, iterable, fields):
        return [e.extract(fields) for e in iterable]

    @classmethod
    def from_dotted(cls, dotkey, val):
        # 'a.b.c', 100 -> {a:{b:{c:100}}}
        # 'a.b.1', 100 -> {a:{b:[None,100]}}

        key, _, sufix = dotkey.partition('.')

        if not sufix:
            if key.isdigit():
                _lst = [None]*int(key) + [val]
                return _lst
            else:
                return cls({key:val})

        if key.isdigit():
            _lst = [None]*int(key) + [cls.from_dotted(sufix, val)]
            return _lst
        else:
            return cls({key: cls.from_dotted(sufix, val)})

    def __init__(self, *arg, **kw):
        super().__init__(*arg, **kw)

        #recursively convert dicts to slovar if they arent already
        for key, val in self.items():
            if isinstance(val, slovar):
                continue
            elif isinstance(val, dict):
                self[key] = slovar(val)
            elif isinstance(val, list):
                new_list = []
                for each in val:
                    if isinstance(each, dict) and not isinstance(each, slovar):
                        new_list.append(slovar(each))
                    else:
                        new_list.append(each)
                self[key] = new_list

    def __call__(self, key):
        return self.extract(key)

    def bad_value_error_klass(self, e):
        return ValueError(e)

    def missing_key_error_klass(self, e):
        return AttributeError(e)

    def __getattr__(self, key):
        if key.startswith('__'):
            raise AttributeError('Attribute error %s' % key)
        try:
            return self[key]
        except KeyError as e:
            raise self.missing_key_error_klass(e.args)

    def __setattr__(self, key, val):
        if key.startswith('__'):
            return super().__setattr__(key,val)

        #this makes a.b.c access work for nested slovars
        if isinstance(val, dict) and not isinstance(val, slovar):
            val = slovar(val)
        self[key] = val

    def __delattr__(self, key):
        self.pop(key, None)

    def __contains__(self, item):
        if isinstance(item, (tuple, list, set)):
            return bool(set(self.keys()) & set(item))
        else:
            return super(slovar, self).__contains__(item)

    def __add__(self, item):
        return self.copy().update(item)

    def __iadd__(self, item):
        return self.update(item)

    def __getitem__(self, key):
        return super(slovar, self).__getitem__(key)

    def to_dict(self, fields=None):
        return self.extract(fields)

    def to_dict_type(self):
        return super(slovar, self).copy()

    @classmethod
    def to(cls, dct):
        if isinstance(dct, slovar):
            return dct.copy()

        return slovar(dct)

    def copy(self):
        return copy.deepcopy(self)

    def deepcopy(self):
        return copy.deepcopy(self)

    def tcast(self, key, val, trs):

        if val is None and not TCAST_NONE:
            log.debug('extracted key %r is None' % key)
            return val

        prev_tr = None

        def concat(val, sep=''):
            SEP = slovar(
                COMMA=',',
                SPACE=' ',
            )

            sep = SEP.get(sep, sep)

            if isinstance(val, list):
                return sep.join([str(it) for it in val if it])
            else:
                return str(val)

        for tr in trs:
            try:
                if 'safe' == tr or 'safe_none' == tr:
                    continue

                elif tr in ('str', 'unicode'):
                    val = str(val)

                elif tr == 'int':
                    val = int(val) if val else val

                elif tr == 'float':
                    val = float(val) if val else val

                elif tr == 'bool':
                    val = bool(val) if val else val

                elif tr == 'flat' and isinstance(val, slovar):
                    val = val.flat()

                elif tr == 'flatall' and isinstance(val, slovar):
                    val = val.flat(keep_lists=False)

                elif tr == 'unflat' and isinstance(val, slovar):
                    val = val.unflat()

                elif tr == 'dt':
                    if val:
                        val = str2dt(val)

                elif tr == 'tm2dt':
                    if val:
                        val = datetime.utcfromtimestamp(val/1000).strftime('%Y-%m-%d %H:%M:%S')

                elif tr == 'dtob':
                    if val:
                        val = ObjectId(val).generation_time

                elif tr == 'strip':
                    if isinstance(val, str):
                        val = val.strip()

                # elif tr == 'ld2dl' and isinstance(val, list):
                #     val = ld2dl(val)

                elif tr in TCAST_FUNCS:
                    prev_tr = tr

                elif prev_tr:
                    if prev_tr == 'sort':
                        reverse = False
                        if tr[0] in ['-', '+']:
                            reverse = tr[0] == '-'
                            tr = tr[1:]
                        val = sort_list(val, tr, reverse = reverse)

                    elif prev_tr == 'index':
                        val = val[int(tr)]

                    elif prev_tr == 'concat':
                        val = concat(val, tr or '')

                    elif prev_tr == 'slice':
                        val = val[:int(tr)]

                    elif prev_tr == 'ld2l':
                        val = ld2l(val, tr)

                    elif prev_tr == 'split':
                        if isinstance(val, str):
                            val = val.split(tr)

                    prev_tr = None

                elif tr.startswith('@'):
                    tr = tr[1:]
                    val = getattr(builtins, tr)(val)

                else:
                    _type = type(val)
                    try:
                        method = getattr(_type, tr)
                        if not isinstance(method, collections.Callable):
                            raise self.bad_value_error_klass('`%s` is not a callable for type `%s`' % (tr, _type))
                        val = method(val)
                    except AttributeError as e:
                        raise self.bad_value_error_klass('type `%s` does not have a method `%s`' % (_type, tr))

            except:
                import sys
                msg = 'typecast failed for key=`%s`, value=`%s`: %s' % (key, val, sys.exc_info()[1])
                log.error(msg)

                if 'safe' in trs:
                    return val
                elif 'safe_none' in trs:
                    return None
                else:
                    raise self.bad_value_error_klass(msg)

        return val

    def extract(self, fields, defaults=None):

        if not fields:
            return self

        op = process_fields(fields)

        def process_assignments(_d):
            for kk, vv in op.assignments.items():
                op.transforms.pop(kk, None)

                val, _, tr = vv.partition(':')

                if val == '__NOW__':
                    val = datetime.utcnow()
                elif val == '__TODAY__':
                    val = datetime.today()
                elif val == '__OID__':
                    val = str(ObjectId())
                elif val == '__NULL__':
                    val = {}

                if '..' in kk:
                    list_key, new_key = kk.split('..')
                    new_val = []

                    for it in _d[list_key]:
                        if tr:
                            val = self.tcast(list_key, val, split_strip(tr, '|'))

                        new_val.append(it.update({new_key:val}))

                    if new_val:
                        _d[list_key] = new_val

                else:
                    trs = split_strip(tr, '|')
                    if 'default' in trs:
                        trs.remove('default')
                        if kk in _d:
                            continue

                    _d[kk] = self.tcast(kk, val, trs) if trs else val

            return _d

        def process_show_as(_d):
            if not op.show_as_r:
                return _d

            _d_show_as = slovar()

            for new_key, key in list(op.show_as_r.items()):
                try:
                    _d_show_as[new_key] = _d.nested_get(key)
                except (KeyError,IndexError):
                    pass

            _d_remainder = slovar()
            for kk, count in collections.Counter(op.exp_only).items():
                if kk in op.show_as:
                    if len(op.show_as[kk]) == count:
                        if op.star:
                            _d = _d.nested_pop(kk)
                        continue

                if not op.star:
                    _d_remainder = _d_remainder.update_with(_d.subset(kk), flatten=kk)

            _d_show_as.update(_d_remainder)

            if op.star:
                return _d_show_as.merge(_d)

            return _d_show_as

        def process_trans(_d):
            for key, trs in list(op.transforms.items()):
                if key in _d:
                    _d[key] = self.tcast(key, _d.get(key), trs)
            return _d

        def process_flats(_d):
            for fld, keep in op.flats.items():
                _d = _d.flat([fld], keep_lists=keep)

            return _d

        def process_unflats(_d):
            if op.unflats:
                _d = _d.unflat(op.unflats)

            return _d

        def process_defaults(_d):
            if defaults:
                _d = _d.update_with(defaults, overwrite=False)

            return _d

        def process_envelope(_d):
            if op.envelope:
                _d = slovar({op.envelope:_d})

            return _d

        _d = self._subset(op)
        _d = process_flats(_d)
        _d = process_show_as(_d)
        _d = process_assignments(_d)
        _d = process_trans(_d)
        _d = process_unflats(_d)
        _d = process_defaults(_d)
        _d = process_envelope(_d)

        return _d

    def get_by_prefix(self, prefix):
        if not isinstance(prefix, list):
            prefixes = [prefix]
        else:
            prefixes = prefix

        _d = slovar()
        for kk, vv in list(self.items()):
            for pref in prefixes:
                _pref = pref[:-1]

                if pref.endswith('*'):
                    if kk.startswith(_pref):
                        ix = _pref.rfind('.')
                        if ix > 0:
                            _pref = _pref[:ix]
                            kk = kk[len(_pref)+1:]
                        _d[kk] = vv
                else:
                    if kk == pref:
                        ix = _pref.rfind('.')
                        if ix > 0:
                            _pref = _pref[:ix]
                            kk = kk[len(_pref)+1:]

                        _d[kk]=vv

        return _d

    def _subset(self, op):
        _d = slovar()

        if op.star:
            _d = self.copy()

        exp_only = op.exp_only[:]
        only_nested_flds = []

        if exp_only:
            for fld in exp_only:
                if fld in self:
                    _d[fld] = self[fld]
                else:
                    try:
                        if fld.endswith('.*'):
                            val = self.nested_get(fld)
                            if isinstance(val, dict):
                                # checking if val keys will overwrite the self
                                # conflict = self.set_keys() & val.set_keys()
                                # if conflict and op.star:
                                #     raise self.missing_key_error_klass(
                                #         'conflict with `%s` field. Reducing to a existing name' % conflict)
                                op.exp_only.extend(val.keys())
                                _d.update(val)
                            else:
                                raise self.bad_value_error_klass('%s must be dict, got %s' % (val, type(val)))
                        elif fld.endswith('*'):
                            _d.update(self.get_by_prefix(fld))
                        else:
                            _d[fld] = self.nested_get(fld)

                        only_nested_flds.append(fld)
                    except (KeyError, IndexError):
                        pass

        if op.exclude:
            _d = _d or self
            _d = _d.remove(op.exclude)

        if only_nested_flds and isinstance(_d, dict):
            _d = _d.unflat(only_nested_flds)

        return _d

    def subset(self, keys, defaults={}):
        if not keys:
            return slovar()

        return self._subset(process_fields(keys)).merge_with(defaults)

    def remove(self, keys, flat=False):

        def _remove_branch(path, _d):
            for key in list(_d.keys()):
                if key.startswith(path):
                    _d = _d.pop(key)

        if isinstance(keys, str):
            keys = [keys]

        _self = self.flat() if flat else self.copy()

        for k in keys:
            if '*' in k:
                _remove_branch(k.replace('*', ''), _self)
            else:
                _self.pop(k, None)

        return _self.unflat() if flat else _self

    def update(self, d_):
        super(slovar, self).update(slovar(d_))
        return self

    def merge(self, d_):
        merge(self, d_)
        return self

    def pop_by_values(self, vals):
        if not isinstance(vals, list):
            vals = [vals]

        for k, v in list(self.items()):
            if v in vals:
                self.pop(k)
        return self

    def nested_get(self, fld):
        _d = self.copy()

        if fld in _d or '.' not in fld:
            return _d[fld]

        for kk in fld.split('.'):
            if '*' == kk:
                break
            # if its a list then access it by index
            if isinstance(_d, list):
                if kk.isdigit():
                    kk = int(kk)
                else:
                    val_lst = []
                    for it in _d:
                        if isinstance(it, slovar):
                           val_lst.append(it.nested_get(kk))
                    return val_lst

            _d = _d[kk]

        return _d

    def nested_in(self, fld):
        if fld in self:
            return True
        elif '.' not in fld:
            return False

        inner = self
        for kk in fld.split('.'):
            if isinstance(inner, dict) and kk in inner:
                inner = inner[kk]
            else:
                return False

        return True

    def nested_pop(self, flds):
        self_d = self.copy()

        if isinstance(flds, str):
            flds = [flds]

        def _pop(fld):
            if fld in self or '.' not in fld:
                self_d.pop(fld)
                return self_d

            parts = fld.split('.')
            _d = self_d

            for kk in parts[:-1]:
                if isinstance(_d, list):
                    kk = int(kk)
                _d = _d[kk]

            try:
                kk = parts[-1]
                if isinstance(_d, list):
                    kk = int(kk)
                _d.pop(kk)

            except (KeyError, IndexError) as e:
                pass

        for fld in flds:
            _pop(fld)

        return self_d

    def get_tree(self, prefix, defaults={}, sep='.'):
        if prefix[-1] != '.':
            prefix += sep

        _dict = slovar(defaults)
        for key, val in list(self.items()):
            if key.startswith(prefix):
                _k = key.partition(prefix)[-1]
                _dict[_k] = val
        return _dict

    def mget(self, keys):
        _d = slovar()
        for it in self:
            if it in keys:
                _d[it] = self[it]
        return _d

    def mpop(self, keys):
        for it in keys:
            self.pop(it, None)

    def has(self, keys, check_type=str,
            err='', _all=True, allow_missing=False,
            allowed_values=[], forbidden_values=[]):
        errors = []

        if isinstance(keys, str):
            keys = [keys]

        self_flat = self.flat().update(self) # update with self to include high level keys too

        def missing_key_error(_type, key):
            if _type == dict:
                missing = ['%s.%s' % (key, val) for val in allowed_values]
            else:
                missing = allowed_values

            return 'Missing key or invalid values for `%s`. Allowed values are: `%s`'\
                                          % (key, missing)

        def error_msg(msg):
            if "%s" in err:
                error = err % msg
            elif err:
                error = err
            else:
                error = msg

            errors.append(error)
            return error

        for key in keys:
            if key in self_flat:
                if check_type and not isinstance(self_flat[key], check_type):
                    error_msg('`%s` must be type `%s`, got `%s` instead'\
                                          % (key, check_type.__name__,
                                             type(self_flat[key]).__name__))

                if allowed_values and self_flat[key] not in allowed_values:
                    error_msg(missing_key_error(check_type, key))

                if self_flat[key] in forbidden_values:
                    error_msg('`%s`=`%s` value is not allowed' % (key, self_flat[key]))

            elif not allow_missing:
                if allowed_values:
                    error_msg(missing_key_error(check_type, key))
                else:
                    error_msg('Missing key: `%s`' % key)

        if (errors and _all) or (not _all and len(errors) >= len(keys)):
            raise self.bad_value_error_klass('.'.join(errors))

        return True

    def transform(self, rules):
        _d = slovar()
        flat_dict = self.flat()

        for path, val in list(flat_dict.items()):
            if path in rules:
                _d.merge(slovar.from_dotted(rules[path], val))

        return _d

    def flat_keys(self, keys, keep_lists=True, sep='.'):
        self_ = self.copy()
        for key in keys:
            val = self_.subset(key)
            if val and isinstance(val, dict):
                self_ = self_.nested_pop(key)
                self_.update(val.flat(keep_lists=keep_lists, sep=sep))

        return self_

    def flat(self, keys=[], keep_lists=True, sep='.'):
        if keys:
            return self.flat_keys(keys, keep_lists=keep_lists, sep=sep)

        return slovar(flat(self, keep_lists=keep_lists, sep=sep))

    def unflat(self, only=[]):
        return slovar(unflat(self, only))

    def set_default(self, name, val):
        cls = slovar
        if name not in self.flat():
            self.merge(cls.from_dotted(name, val))
        return val

    def with_defaults(self, **defaults):
        return self.update_with(defaults, overwrite=False)

    def fget(self, key, *arg, **kw):
        return self.flat().get(key, *arg, **kw)

    def update_with(self, _dict, overwrite=True, append_to=None,
                                                 append_to_set=None,
                                                 flatten=None,
                                                 merge_to=None,
                                                 remove_from=None):

        self_dict = self.copy()
        if not _dict:
            return self_dict

        def process_append_to_param(_lst):
            if isinstance(_lst, str):
                _lst = [_lst]

            if not _lst:
                return {}

            _d = {}
            for each in (_lst or []):
                k,_,sk = each.partition(':')
                _d[k]=sk

                #reference to nested field?
                if '.' in k and (not flatten or (isinstance(flatten, list) and k.split('.')[0] not in flatten)):
                    raise self.bad_value_error_klass(
                        'list operation referrers to nested field `%s` without flattening.'
                        ' forgot to pass `flatten=%s`?' % (k, k.split('.')[0]))
            return _d

        append_to = process_append_to_param(append_to)
        append_to_set = process_append_to_param(append_to_set)
        merge_to = process_append_to_param(merge_to)
        remove_from = process_append_to_param(remove_from)

        def _build_list(_lst, new_val):
            if isinstance(_lst, list):
                if isinstance(new_val, list):
                    _lst.extend(new_val)
                else:
                    _lst.append(new_val)
            else:
                raise self.bad_value_error_klass('`%s` is not a list' % key)

            return _lst

        def _append_to(self_dict, key, val):
            _lst = _build_list(self_dict.get(key, []), val)
            sort_key = append_to.get(key)
            sort_method = None
            reverse = False

            if sort_key:
                if sort_key.startswith('-'):
                    sort_key = sort_key[1:]
                    reverse = True
                elif sort_key.startswith('+'):
                    sort_key = sort_key[1:]

                if sort_key:
                    sort_method = lambda x: x.get(sort_key)

                _lst = sorted(_lst, key=sort_method, reverse=reverse)

            return _lst

        def _append_to_set(self_dict, key, val):
            new_lst = _build_list(self_dict.get(key, []), val)
            set_key = append_to_set.get(key)

            if set_key.startswith('-'):
                reverse_order = True
                set_key = set_key[1:]
            else:
                reverse_order = False

            #ie append_to_set=people:full_name. `full_name` is a set_key
            #this will mean make people unique for inner field `full_name`
            if set_key:
                _uniques = []
                _met = []
                _not_found = []

                #reverse the list so new values overwrite old ones,
                #since it was appended at the end
                for each in reversed(new_lst):
                    if not each:
                        log.debug('Empty item in the `%s:%s` list.Skip.', key, set_key)
                        continue

                    if set_key not in each:
                        _not_found.append(each)
                        continue

                    if each[set_key] in _met:
                        continue

                    _met.append(each[set_key])
                    _uniques.append(each)

                new_lst = sorted(_uniques, key= lambda x: x.get(set_key), reverse=reverse_order)
                new_lst.extend(_not_found)
            else:
                try:
                    new_lst = list(set(new_lst))
                except TypeError as e:
                    raise self.bad_value_error_klass('items in `%s` list not hashable. missed the set_key ?'\
                                     % (key))

            return new_lst

        def _merge_to(self_dict, key, val):
            set_key = merge_to.get(key)
            new_lst = []
            new_keys = []

            if not set_key:
                raise self.bad_value_error_klass('merge_to must contain a set key')

            if not isinstance(val, list):
                val = [val]

            for each in self_dict.get(key, []):
                if each.get(set_key):
                    for it in val:
                        if not it.get(set_key):
                            continue

                        if it[set_key] == each[set_key]:
                            each = each.update_with(it)

                        new_keys.append(it[set_key])

                new_lst.append(each)

            return new_lst

        def _remove_from(self_dict, key, val):
            set_key = remove_from.get(key)
            new_lst = self_dict.get(key, [])[:] # copy

            for vv in val:
                for vvv in new_lst:
                    if set_key:
                        if not isinstance(vv, dict) or not isinstance(vvv, dict):
                            raise self.bad_value_error_klass(
                                'set key `%s` can be specified only for dicts. Got : `%s` and `%s`' % (
                                                                                        set_key, vv, vvv))

                        if set_key in vv and vv[set_key] == vvv.get(set_key, None):
                            new_lst.remove(vvv)
                    else:
                        if vv == vvv:
                            new_lst.remove(vvv)

            return new_lst

        def can_overwrite(key, overwrite, flat_overwrites):
            if overwrite == True:
                return True

            if isinstance(overwrite, list):
                if flatten == True:
                    return True

                for it in flat_overwrites:
                    if key.startswith('%s.'%it):
                        return True

                if key in overwrite:
                    return True

            return False

        flat_overwrites = set()
        if flatten:
            flat_keys = flatten if isinstance(flatten, list) else None
            self_dict = self_dict.flat(keys=flat_keys)
            _dict = _dict.flat(keys=flat_keys)

            if flat_keys and isinstance(overwrite, list):
                s_overwrites = set(overwrite)
                flat_overwrites = (set(flat_keys) & s_overwrites)
                overwrite = list(s_overwrites - flat_overwrites)

        for key, val in list(_dict.items()):

            if key in append_to:
                self_dict[key] = _append_to(self_dict, key, val)
            elif key in append_to_set:
                self_dict[key] = _append_to_set(self_dict, key, val)
            elif key in merge_to:
                self_dict[key] = _merge_to(self_dict, key, val)
            elif key in remove_from:
                self_dict[key] = _remove_from(self_dict, key, val)
            elif key not in self_dict:
                self_dict[key] = val
            elif overwrite and can_overwrite(key, overwrite, flat_overwrites):
                self_dict[key] = val

        if flatten:
            self_dict = self_dict.unflat()

        return self_dict

    def merge_with(self, _dict):
        return self.update_with(_dict, overwrite=False)

    def contains(self, other, exclude=None):
        other_ = other.subset(exclude)
        return not other_ or self.subset(list(other_.keys())) == other_

    def pop_many(self, keys):
        if not keys:
            return self.copy()

        poped = slovar()
        pop_keys = self.extract(keys).keys()
        for key in pop_keys:
            poped[key] = self.pop(key, None)
        return poped

    def any_key(self, keys):
        if not keys:
            return False
        return any(name in self for name in keys)

    def all_keys(self, keys):
        if not keys:
            return False
        return all(name in self for name in keys)

    def key_diff(self, keys):
        return list(set(self.keys()) - set(keys))

    def key_union(self, keys):
        return list(set(self.keys()) | set(keys))

    def key_intersection(self, keys):
        return list(set(self.keys()) & set(keys))

    def diff(self, sl2, diff_fields=[], flat_keys=[]):

        _self = self

        if flat_keys:
            _self = _self.flat(keys=flat_keys)
            sl2 = sl2.flat(keys=flat_keys)

        self_diff = slovar()
        sl2_diff = slovar()
        _diff = slovar()

        if diff_fields:
            diff_fields = list(_self.extract(diff_fields).keys())

        for kk in (diff_fields or list(_self.keys())):
            selfv = _self.get(kk, None)
            sl2v = sl2.get(kk, None)

            if selfv != sl2v:
                _diff[kk] = {'from': selfv, 'to': sl2v}
                self_diff[kk] = selfv
                sl2_diff[kk] = sl2v

        if not diff_fields:
            #iterate through missing keys in _self
            for kk in set(sl2.keys())-set(_self.keys()):
                _diff[kk] = {'from': None, 'to': sl2[kk]}
                self_diff[kk] = None
                sl2_diff[kk] = sl2[kk]

        return self_diff, sl2_diff

    def add_to_list(self, list_name, items, unique=False, position=None, sort_key=None):
        _self = self

        if list_name not in _self:
            _self[list_name] = []

        if not items:
            return _self

        if not isinstance(items, list):
            items = [items]

        if position is not None:
            for ix in range(len(items)):
                _self[list_name].insert(position+ix, items[ix])
        else:
            _self[list_name].extend(items)

        if unique:
            _self[list_name] = list(set(_self[list_name]))

        if sort_key:
            _self[list_name] = sorted(_self[list_name], key=sort_key)

        return _self

    def concat_values(self, sep=':'):
        concat = []

        _keys = sorted(self.keys())

        for kk in _keys:
            concat.append(str(self[kk]))

        return sep.join(_keys), sep.join(concat)

    def ordered_values(self, keys):
        return [self.get(kk) for kk in keys if kk in self]

    def call_converter(self, name, *arg, **kw):
        try:
            return getattr(convert, name)(self, *arg, **kw)
        except KeyError as e:
            raise self.missing_key_error_klass(e)
        except ValueError as e:
            raise self.bad_value_error_klass(e)

    def asbool(self, *arg, **kw):
        return self.call_converter('asbool', *arg, **kw)

    def aslist(self, *arg, **kw):
        return self.call_converter('aslist', *arg, **kw)

    def asset(self, *arg, **kw):
        return self.call_converter('asset', *arg, **kw)

    def asint(self, *arg, **kw):
        return self.call_converter('asint', *arg, **kw)

    def asfloat(self, *arg, **kw):
        return self.call_converter('asfloat', *arg, **kw)

    def asdict(self, *arg, **kw):
        return self.call_converter('asdict', *arg, **kw)

    def asdt(self, *arg, **kw):
        return self.call_converter('asdt', *arg, **kw)

    def asstr(self, *arg, **kw):
        return self.call_converter('asstr', *arg, **kw)
        return asstr(self, *arg, **kw)

    def asrange(self, *arg, **kw):
        return self.call_converter('asrange', *arg, **kw)

    def asqs(self, *arg, **kw):
        return self.call_converter('asqs', *arg, **kw)

    def asdtob(self, *arg, **kw):
        return self.call_converter('asdtob', *arg, **kw)

    def json(self):
        return json_dumps(self)

    def set_keys(self):
        #useful for testing mainly
        return set(self.keys())



import re
from datetime import datetime
from dateutil import parser as dt_parser, relativedelta as dt_relativedelta
import logging

log = logging.getLogger(__name__)

def dot_split(s):
    return [part for part in re.split(r"(?<!\.)\.(?!\.)", s)]


def split_strip(_str, on=',', remove_empty=True):
    lst = (_str if isinstance(_str, list) else _str.split(on))
    lst = [e.strip() for e in lst]
    if remove_empty:
        lst = list(filter(bool, lst))
    return lst


def str2rdt(strdt):
    matches = dict(
        s = 'seconds',
        m = 'minutes',
        h = 'hours',
        d = 'days',
        M = 'months',
        y = 'years'
    )

    # is it a relative date ?
    rg = re.compile('(([-+]?)(\d+))([smhdMy])\\b', re.DOTALL)
    m = rg.search(strdt)
    if m:
        number = int(m.group(1))
        word = m.group(4)
        if word in matches:
            log.debug('relative date detected: %s', {matches[word]:number})
            return dt_relativedelta.relativedelta(**{matches[word]:number})


def str2dt(strdt, _raise=False):
    if not strdt:
        raise ValueError('Datetime string can not be empty or None')

    if isinstance(strdt, datetime):
        return strdt

    dt = str2rdt(strdt)
    if dt:
        return datetime.utcnow()+dt
    try:
        return dt_parser.parse(strdt)
    except ValueError as e:
        msg = 'Datetime string `%s` not recognized as datetime. Did you miss +- signs for relative dates?' % strdt
        if _raise:
            raise ValueError(msg)
        else:
            log.error(msg)


def snake2camel(text):
    '''turn the snake case to camel case: snake_camel -> SnakeCamel'''
    return ''.join([a.title() for a in text.split('_')])


def camel2snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()





class SlovarKeyError(KeyError):
    pass


class SlovarValueError(ValueError):
    pass


def set_exceptions(key_error, value_error):
    global SlovarKeyError, SlovarValueError
    SlovarKeyError = key_error
    SlovarValueError = value_error



import time
import json
import pytest
from slovar import slovar


LOREM = 'Lorem'


class TestSlovarComplex(object):
    sample_d = slovar(
        a=1,
        b=LOREM,
        c=list(range(20)),
        d={
            'i': [LOREM] * 10,
            'ii': {
                'aa': LOREM,
                'ab': [5, 6, [7, 8, 9]],
            },
            'iii': [1,2,3,4],
            'iiii': [
                {'j': LOREM},
                {'jj': LOREM},
            ]
        },
        f=[[1, 2, 3], 4, 5],
        g=[1, 2, 4, [4, 5], 6],
        h=[1, 2, 4, [[4, 5], 6]],
    )

    def test_flat(self):
        d = slovar(self.sample_d)
        assert slovar(d.flat()).unflat() == self.sample_d

    def test_flat_lists(self):
        d = slovar(self.sample_d)
        assert slovar(d.flat(keep_lists=False)).unflat() == self.sample_d

    def test_unflat(self):
        d = slovar(self.sample_d).flat()
        assert d.unflat() == self.sample_d

    def test_unflat_lists(self):
        d = slovar(self.sample_d).flat(keep_lists=False)
        assert d.unflat() == self.sample_d

    def test_extract(self):
        d = slovar(self.sample_d)
        args = ['a', 'b', 'c']
        assert d.extract(args) == {
            'a': self.sample_d['a'],
            'b': self.sample_d['b'],
            'c': self.sample_d['c'],
        }

    def test_extract_with_defaults(self):
        _d = slovar(a=1,b=2)
        #adds the default
        assert _d.extract('a,d', defaults={'d':3}) == {'a':1, 'd':3}
        #does not overwrite with default if exists
        assert _d.extract('a,b', defaults={'b':3}) == {'a':1, 'b':2}

    def test_extract_nested(self):
        dd = slovar(self.sample_d)

        #get nested dict
        _d = dd.extract(['a', 'b', 'c', 'd.ii'])
        assert set(_d.flat().keys()) == set(['a', 'b', 'c', 'd.ii.aa', 'd.ii.ab'])

        #get nested list
        _d = dd.extract(['a', 'b', 'c', 'd.ii.ab'])
        assert set(_d.flat().keys()) == set(['a', 'b', 'c', 'd.ii.ab'])

        #get nested list item
        _d = dd.extract(['d.ii.ab.2'])
        assert set(_d.flat().keys()) == set(['d.ii.ab'])

    def test_extract_exclude(self):
        d = slovar(self.sample_d)
        args = ['-a', '-b', '-c', '-e', '-g', '-h']
        assert d.extract(args) == {
            'd': self.sample_d['d'],
            'f': self.sample_d['f'],
        }

    def test_subset(self):
        d = slovar(self.sample_d)
        args = ['a', 'b', 'd']
        assert d.subset(args) == {
            'a': self.sample_d['a'],
            'b': self.sample_d['b'],
            'd': self.sample_d['d'],
        }

    def test_subset_exclude(self):
        d = slovar(self.sample_d)
        args = ['-a', '-b', '-c', '-e', '-g', '-h']
        assert d.subset(args) == {
            'd': self.sample_d['d'],
            'f': self.sample_d['f'],
        }

    def test_update_with(self):
        # Include d in both to have a collision
        d = slovar(self.sample_d).subset(['a', 'b', 'd', 'g', 'h'])
        e = slovar(self.sample_d).subset(['c', 'd', 'e', 'f'])
        assert d.update_with(e) == self.sample_d

    def test_update_with_append_to(self):
        # Include d in both to have a collision
        d = slovar(self.sample_d).subset(['a', 'b', 'd', 'g', 'h'])
        e = slovar(self.sample_d).subset(['c', 'd', 'e', 'f'])
        a = []
        assert d.update_with(e, append_to=a) == self.sample_d

        d1 = slovar(
            a = [],
            aa = [1],
            aaa = [{'b':1, 'c':2}]
        )
        d2 = slovar(
            a = 1,
            aa = 2,
            aaa= [{'b':3, 'c': 3}, {'b':2, 'c': 33}]
        )

        d3 = d1.update_with(d2, append_to='aaa')
        assert d3.aaa == [{'b':1, 'c':2}, {'b':3, 'c': 3}, {'b':2, 'c': 33}]

        d3 = d1.update_with(d2, append_to='aaa:b')
        assert d3.aaa == [{'b':1, 'c':2}, {'b':2, 'c': 33}, {'b':3, 'c': 3}]

        d3 = d1.update_with(d2, append_to='aaa:-b')
        assert d3.aaa == [{'b':3, 'c': 3}, {'b':2, 'c': 33}, {'b':1, 'c':2}]

    def test_update_with_append_to_set(self):
        d1 = slovar(
            a = [],
            aa = [1],
            aaa = [{'b':1, 'c':2}]
        )
        d2 = slovar(
            a = 1,
            aa = 2,
            aaa= [{'b':1, 'c': 3}, {'b':2, 'c': 33}]
        )

        for ix in range(2):
            d3 = d1.update_with(d2, append_to_set=[
                    'a',
                    'aa',
                    'aaa:b'
            ])

            assert len(d3.a) == 1
            assert len(d3.aa) == 2
            assert len(d3.aaa) == 2

        d3 = d1.update_with({'x':1}, append_to='x')
        assert 'x' in d3

    def test_update_with_append_to_set2(self):
        d1 = slovar(
            a = [],
            aa = [1],
            aaa = [{'b':1, 'c':2}]
        )

        d3 = d1.update_with({'aaa':{'b': 1, 'c':11, 'd':33}}, append_to_set=['aaa:b'])
        assert d3.aaa[0]['c'] == 11
        assert d3.aaa[0]['d'] == 33

    def test_update_with_append_to_set_order(self):
        d1 = slovar(
            a = [
                dict(b = 10),
                dict(b = 20),
                dict(b = 30),
            ]
        )

        d3 = d1.update_with(dict(a=[dict(b=5), dict(b=15)]), append_to_set='a:-b')
        # assert d3.a == [{'b': 30}, {'b': 20}, {'b': 15}, {'b': 10}, {'b': 5}]

        d3 = d1.update_with(dict(a=[dict(b=5), dict(b=15)]), append_to_set='a:b')
        # assert d3.a == [{'b': 30}, {'b': 20}, {'b': 15}, {'b': 10}, {'b': 5}]

    def test_update_with_overwrite(self):
        d1 = slovar(
            a = {'b': 1},
            c = 2,
            d = [2]
        )

        d2 = d1.update_with({'c':1, 'd':3})
        assert d2.c == 1
        assert d2.d == 3

        d2 = d1.update_with({'c':1, 'd':3}, overwrite=False)
        assert d2.c == 2
        assert d2.d == [2]

        d2 = d1.update_with({'d':1}, append_to='d')
        assert d2.d == [2,1]

        d2 = d1.update_with({'d':4}, append_to='d', overwrite=False)
        assert d2.d == [2,4]

    def test_update_with_flat(self):

        d1 = slovar(
            a = {'b':
                    {'c': 1},
                 'bb': 2},
        )

        d2 = slovar(
            a = {'b':
                    {'d': 2}}
        )

        d3 = d1.update_with(d2, flatten=True)
        assert d3.a.b.d == 2
        assert d3.a.b.c == 1

    def test_extract_with_assignment(self):
        assert slovar(a=1).extract('a:=2') == slovar(a='2')

    def test_update_with_flat_keys(self):
        d1 = slovar(
            a = slovar(b=1),
            c = slovar(d=2),
            e = 3
        )

        assert d1.flat() == slovar({'a.b':1, 'c.d':2, 'e':3})
        assert 'a' not in d1.flat()
        assert d1.flat(['a']) == slovar({'a.b':1, 'c':slovar(d=2), 'e':3})
        assert d1.flat(['a', 'e']) == slovar({'a.b':1, 'c':slovar(d=2), 'e':3})
        assert d1.flat(['e']) == d1

        d2 = d1.update_with(slovar(a=slovar(bb=4)))
        assert 'b' not in d2.a
        assert d2.a.bb == 4
        assert d2.c.d == 2

        d2 = d1.update_with(slovar(a=slovar(bb=4), c=slovar(ff=5)), flatten=True)
        assert 'b' in d2.a
        assert d2.a.bb == 4
        assert 'd' in d2.c
        assert d2.c.ff == 5

        #flatten only `a` and `c` should be overwritten
        d2 = d1.update_with(slovar(a=slovar(bb=4), c=slovar(ff=5)), flatten=['a'])
        assert 'b' in d2.a
        assert d2.a.bb == 4
        assert 'd' not in d2.c
        assert d2.c.ff == 5

    def test_extract_envelope(self):
        d1 = slovar(
            a = 1,
            b = slovar(c = 2)
        )

        # d2 = d1.extract('__as__d')
        # assert 'd' in d2
        # assert d2.d == d1
        # assert 'a' not in d2
        # assert 'b' not in d2
        # assert 'd.a' in d2.flat()
        # assert 'd.b.c' in d2.flat()

        d2 = d1.extract('a,__as__d')
        assert d2.d.set_keys() == set(['a'])

    def test_extract_assign(self):
        d1 = slovar(
            c = '123',
            b='345'
        )

        d2 = d1.extract('c__as__a.c:float,b__as__a.b:float,a.dd:=dd').unflat()
        assert d2.a.c == 123
        assert d2.a.b == 345
        assert d2.a.dd == 'dd'

    def test_update_with_merge_to(self):
        d1 = slovar(
            a = [
                dict(b=1, c=2),
                dict(b=11, c=22),
                dict(b=11, c=33),
                dict(c=44),
            ]
        )

        d1.update_with(slovar(a=1), merge_to='x')
        d1.update_with(slovar(), merge_to='a:b')

        with pytest.raises(Exception):
            d1.update_with(slovar(a=1), merge_to='a')

        #should not change d1 since there is not b=55 in it
        d2 = slovar(a={'b':55, 'c': 222})
        d3 = d1.update_with(d2, merge_to='a:b')
        assert d3 == d1

        d2 = slovar(a=[{'b':11, 'c': 222, 'd': 444}])
        d3 = d1.update_with(d2, merge_to='a:b')

        assert len(d3) == len(d1)
        assert 'd' not in d1.a

        assert d3.a[1]['c'] == 222
        assert d3.a[1]['d'] == 444

        assert d3.a[2]['c'] == 222
        assert d3.a[2]['d'] == 444

        d3 = d2.update_with(d1, merge_to='a:b')
        assert d3.a[0]['c'] == 33

    def test_unflat_extract(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d':[1,2],
                'e': 1})

        d2 = d1.unflat().extract('a.b')
        assert 'a' in d2
        assert 'a.b' not in d2

        d2 = d1.extract(['a.b'])
        assert 'a.b' in d2
        assert d2['a.b'] == 1
        assert 'a.c' not in d2

        d2 = d1.unflat().extract('a')
        assert d2.set_keys() == set(['a'])

        d2 = d1.extract('a.d')
        assert d2.set_keys() == set(['a.d'])

        d2 = d1.unflat().extract('a.d')
        assert d2.flat().set_keys() == set(['a.d'])

    def test_subset2(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d.d':[1,2],
                'e': 1,
            }).unflat()

        d2 = d1.subset('a.d.d,a.*')
        assert d2.set_keys() == set(['a', 'b', 'c', 'd'])
        assert d2.a.d.d == d2.d.d

    def test_reduce_extract(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d.d':[1,2],
                'e': 1}).unflat()

        d2 = d1.extract('a.*')
        assert d2.set_keys() == set(['b', 'c', 'd'])

        d2 = d1.flat().extract('a.*')
        assert d2.set_keys() == set()

        d2 = d1.extract('a.d.d__as__dd,a.d.*')
        assert d2.flat().set_keys() == set(['dd', 'd'])

    def test_show_as(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d.dd':[1,2],
                'a.d.bb': 'bb'}).unflat()

        d2 = d1.extract('a.d.dd__as__dlist')
        assert d2.set_keys() == set(['dlist'])

        d2 = d1.flat().extract('a.d.dd__as__dlist')
        assert d2.set_keys() == set(['dlist'])

        d2 = d1.extract('a.d.dd__as__dlist, a.d.dd')
        assert d2.set_keys() == set(['dlist', 'a'])

    def test_star(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d.dd':[1,2],
                'e': 1}).unflat()

        d2 = d1.extract('a.d.dd__as__aa,*')
        assert d2.set_keys() == set(['aa', 'a', 'e'])

    def test_flat(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d.dd':[1,2],
                'e': 1}).unflat()

        d2 = d1.flat(['a'])
        assert d2.set_keys() == set(['a.b', 'a.c', 'a.d.dd', 'e'])

    def test_nested_get(self):
        d1 = slovar({
            'e':[{'ee': 1, 'xx': 1}, {'ee': 2, 'yy': 1}]
        }).unflat()

        d2 = d1.nested_get('e.ee')
        assert set(d2) == set([1,2])

    def test_nested_pop(self):
        d1 = slovar({
                'a.b':1,
                'a.c': 2,
                'a.d.dd':[1,2],
                'e': 1}).unflat()

        d2 = d1.nested_pop('a.d.dd.0')
        assert d2.a.d.dd[0] == 2
        assert len(d2.a.d.dd) == 1

        assert 'a.d.dd' in d1.flat()
        d2 = d1.nested_pop('a.d.dd')
        assert 'a.d.dd' not in d2.flat()

        d2 = d1.nested_pop('a.d')
        d2 = d1.nested_pop('a')

        d2 = d1.nested_pop(['a.b', 'e'])
        assert d2.flat().set_keys() == set(['a.c', 'a.d.dd'])

    def test_mix1(self):
        d1 = slovar({
            'a': 1,
            'b': [1,2,3],
            'c.c': '1',
            'c.cc': [11,22,32],
            'd.d.d': 1,
            'd.d.dd': [111,222,333],
            'e':[{'ee': 1}, {'ee': 2}]
        }).unflat()

        fields = ['a', 'b', 'c.c__as__cc:int',
                                  'd.d', 'd.d.dd__as__ddd',
                                  'e.1',
                                  'g.g:=ggg',
                                  'd:unflat'
                                  ]

        d2 = d1.extract(fields, defaults={'f': 1})
        assert d2.set_keys() == set(['a', 'b', 'cc', 'd', 'ddd', 'e', 'f', 'g.g'])

    def test_reducer_with_show_as(self):
        d1 = slovar({
            'a': 1,
            'b': [1,2,3],
            'c.c': '1',
            'c.cc': [11,22,32],
            'd.d.d': 1,
            'd.d.dd': [111,222,333],
            'e':[{'a': {'b': 1, 'c': 2}, 'aa': 11}, {'a': {'b': 2}, 'aa': 22}]
        }).unflat()

        d2 = d1.extract('e.0.*,e.aa__as__aaa')
        assert d2.set_keys() == set(['a', 'aa', 'aaa'])
        assert d2['a'] == {'b': 1, 'c': 2}
        assert d2['aaa'] == [11,22]

    def test_flat_reducer(self):
        # if fields are flat originally, extract will treat them just as normal keys, not nested.
        # so doing nesting ops should not work

        d1 = slovar({
            'c.c': '1',
            'c.cc': [11,22,32],
            'd.d.d': 1,
        })

        d2 = d1.extract('c.*')
        assert not d2

        d2 = d1.extract('c*')
        assert d2.set_keys() == set(['c.c', 'c.cc'])

        d2 = d1.extract('d.d')
        assert not d2

    def test_unflat_show_as(self):
        d1 = slovar({
            'a': '1',
            'b': [11,22,32],
            'd.d': 1,
        })

        d2 = d1.extract(['a__as__aa.a', 'b__as__aa.b', 'aa:unflat', 'd.d'])
        assert d2.set_keys() == set(['aa', 'd'])

    def test_show_as_double(self):
        d1 = slovar({
            'a': '1',
            'b': [11,22,32],
            'd.d': 1,
        })

        d2 = d1.extract(['a__as__b', 'a__as__bb'])
        assert d2.set_keys() == set(['b', 'bb'])


    @pytest.mark.skip('nested trans is not implemented')
    def test_nested_trans(self):
        d1 = slovar({'abc': 1, 'geo': {'lat': '50.420907', 'lon': '9.414015'}})
        d2 = d1.extract('geo.lat:float,geo.lon:float,*')

        assert type(d2.geo.lat) == float
        assert 'abc' in d2

    def test_flat_unflat(self):
        d1 = slovar({
            'a': '1',
            'b': [{'bb': 11},22,32],
            'd.d': 1,
        }).unflat()

        d2 = d1.extract('b.0.bb:=111,b:flatall|unflat')
        assert d2.set_keys() == set(['b'])
        assert d2.b[0]['bb'] == '111'

    def test_as_star(self):
        d1 = slovar({
            'a': '1',
            'b': [{'bb': 11,},22,32],
            'd.d': 1,
        }).unflat()

        d2 = d1.extract('d.d__as__d.ddd,*')
        assert d2.flat().set_keys() == set(['d.ddd', 'a', 'b', 'd'])

    def test_multi_nested_with_show_as(self):
        d1 = slovar({
            'x': 0,
            'a.a': '1',
            'a.b': [{'bb': 11},22,32],
            'a.c': 1,
            'a.d': 'd',
        }).unflat()

        d2 = d1.extract('a.a,a.b,a.c,x__as__xx')
        assert d2.set_keys() == set(['a', 'xx'])
        assert d2.flat().set_keys() == set(['a.a', 'a.b', 'a.c', 'xx'])

    def test_default_assign(self):
        d1 = slovar(a=1,b=2,c=3)

        d2 = d1.extract('a:=10:int')
        assert d2.a == 10

        d2 = d1.extract('a:=10:int|default,b')
        assert d2.a == 1
        assert d2.set_keys() == set(['a', 'b'])

        d2 = d1.extract('d:=100:int|default')
        assert d2.d == 100

        d2 = d1.extract('a:=10:int|default,*')
        assert d2.set_keys() == set(['a', 'b', 'c'])

    def test_update_with_remove_from(self):
        d1 = slovar({
            'a': [],
            'aa.a': [1,2,3],
            'aa.b': [{'a':1}, {'b':2}],
            'aa.c': [{'a': 1}, 2, 3],
        }).unflat()

        d2 = slovar({
            'a': [1],
            'aa.a': [2],
            'aa.b': [{'a':1}, {'x': 1}],
            'aa.c': [3]
        }).unflat()

        d3 = d1.update_with(d2, remove_from='a')
        assert d3.a == d1.a

        d3 = d1.update_with(d2, remove_from='aa.a', flatten=['aa'])
        assert d3.aa.a == [1,3]

        with pytest.raises(ValueError):
            d3 = d1.update_with(d2, remove_from='aa.a:xxx', flatten=['aa'])

        d3 = d1.update_with(d2, remove_from='aa.c', flatten=['aa'])
        assert d3.aa.c == [{'a':1}, 2]

        with pytest.raises(ValueError):
            d3 = d1.update_with(d2, remove_from='aa.c:xxx', flatten=['aa'])

        d3 = d1.update_with(d2, remove_from='aa.b:a', flatten=['aa'])
        assert 'a' not in d3.aa.b

    def test_update_with_nested_list(self):

        d1 = slovar({
            'a': [],
            'aa.a': [1,2,3],
            'aa.b': [{'a':1}, {'b':2}],
            'aa.c': [{'a': 1}, 2, 3],
        }).unflat()

        d2 = slovar({
            'a': [1],
            'aa.a': [4],
            'aa.b': [{'a':1}, {'x': 1}],
            'aa.c': [3]
        }).unflat()

        d3 = d1.update_with(d2, append_to_set='aa.a', flatten='aa.a')
        assert d3.aa.a == [1,2,3,4]

    def test_nested_in(self):
        d1 = slovar({
            'a.b.c': 1,
            'x': 2
            }).unflat()

        assert d1.nested_in('a')
        assert d1.nested_in('a.b')
        assert d1.nested_in('a.b.c')
        assert d1.nested_in('x')
        assert not d1.nested_in('a.y')
        assert not d1.nested_in('x.y')

    def test_simple_combo1(self):
        d1 = slovar({
            'a.b.b1': 1,
            'a.b.b2': 2,
            'd.dd': {}
            }).unflat()

        d2 = d1.extract('d, a.b.b1__as__bb1')
        assert 'd' in d2
        assert 'bb1' in d2


    def test_flat_empty(self):
        d1 = slovar({
            'a.b.b1': 1,
            'a.b.b2': 2,
            'd.dd': {}
            }).unflat()

        d2 = d1.flat()
        assert 'd.dd' in d2




import logging

log = logging.getLogger(__name__)


def resolve(name, module=None):
    """Resole dotted name to python module
    """
    name = name.split('.')
    if not name[0]:
        if module is None:
            raise ValueError('relative name without base module')
        module = module.split('.')
        name.pop(0)
        while not name[0]:
            module.pop()
            name.pop(0)
        name = module + name

    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used += '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)

    return found


def maybe_dotted(module, throw=True):

    def _import(module):
        if isinstance(module, str):
            module, _, cls = module.partition(':')
            module = resolve(module)
            if cls:
                return getattr(module, cls)

        return module

    if throw:
        return _import(module)
    else:
        try:
            return _import(module)
        except ImportError as e:
            log.error('%s not found. %s' % (module, e))


def fuzzy_match(name1, name2):
    import jellyfish

    """
        match_rating_comparision - phonetic comparison. works on 5 chars and more. returns None if < 4 chars
        jaro_distance - string-edit distance.
            values are floats between [0-1], where 0 is completely dissimilar strings and 1 is identical
    """

    n1 = str(name1.upper())
    n2 = str(name2.upper())

    return dict(
            mra = jellyfish.match_rating_comparison(n1, n2),
            jaro_distance = round(jellyfish.jaro_distance(n1, n2), 1)
        )

def fuzzy_sort(n1, nn, cuttoff=0.8):
    """
        will sort nn by the best fuzzy match with n1 using fuzzy_match method in this module.
    """

    matches = []

    for each in nn:
        match = fuzzy_match(n1, each)
        if match['mra'] and match['jaro_distance'] >= cuttoff:
            matches.append([each, match])

    def _key_func(xx):
        return xx[1]['jaro_distance']

    if matches:
        return sorted(matches, key=_key_func, reverse=True)

